package org.apache.bcel.verifier.statics;

package .var.folders.cl.9twcgjfx4vsg5zl73c6wxmvm0000gn.T;

import java.util.Arrays;
import org.apache.bcel.Repository;
import org.apache.bcel.classfile.Constant;
import org.apache.bcel.classfile.ConstantClass;
import org.apache.bcel.classfile.ConstantInterfaceMethodref;
import org.apache.bcel.classfile.ConstantMethodref;
import org.apache.bcel.classfile.ConstantNameAndType;
import org.apache.bcel.classfile.ConstantUtf8;
import org.apache.bcel.classfile.Field;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.generic.ALOAD;
import org.apache.bcel.generic.ANEWARRAY;
import org.apache.bcel.generic.ASTORE;
import org.apache.bcel.generic.ArrayType;
import org.apache.bcel.generic.CHECKCAST;
import org.apache.bcel.generic.ConstantPoolGen;
import org.apache.bcel.generic.DLOAD;
import org.apache.bcel.generic.DSTORE;
import org.apache.bcel.generic.EmptyVisitor;
import org.apache.bcel.generic.FLOAD;
import org.apache.bcel.generic.FSTORE;
import org.apache.bcel.generic.FieldInstruction;
import org.apache.bcel.generic.GETSTATIC;
import org.apache.bcel.generic.IINC;
import org.apache.bcel.generic.ILOAD;
import org.apache.bcel.generic.INSTANCEOF;
import org.apache.bcel.generic.INVOKEDYNAMIC;
import org.apache.bcel.generic.INVOKEINTERFACE;
import org.apache.bcel.generic.INVOKESPECIAL;
import org.apache.bcel.generic.INVOKESTATIC;
import org.apache.bcel.generic.INVOKEVIRTUAL;
import org.apache.bcel.generic.ISTORE;
import org.apache.bcel.generic.Instruction;
import org.apache.bcel.generic.InvokeInstruction;
import org.apache.bcel.generic.LDC;
import org.apache.bcel.generic.LDC2_W;
import org.apache.bcel.generic.LLOAD;
import org.apache.bcel.generic.LOOKUPSWITCH;
import org.apache.bcel.generic.LSTORE;
import org.apache.bcel.generic.LoadClass;
import org.apache.bcel.generic.MULTIANEWARRAY;
import org.apache.bcel.generic.NEW;
import org.apache.bcel.generic.NEWARRAY;
import org.apache.bcel.generic.ObjectType;
import org.apache.bcel.generic.PUTSTATIC;
import org.apache.bcel.generic.RET;
import org.apache.bcel.generic.ReferenceType;
import org.apache.bcel.generic.TABLESWITCH;
import org.apache.bcel.generic.Type;
import org.apache.bcel.verifier.VerificationResult;
import org.apache.bcel.verifier.Verifier;
import org.apache.bcel.verifier.VerifierFactory;
import org.apache.bcel.verifier.exc.AssertionViolatedException;
import org.apache.bcel.verifier.exc.StaticCodeInstructionOperandConstraintException;
import org.apache.bcel.verifier.statics.Pass3aVerifier;

class InstOperandConstraintVisitor extends EmptyVisitor {
  private final ConstantPoolGen constantPoolGen;
  
  InstOperandConstraintVisitor(ConstantPoolGen constantPoolGen) {
    this.constantPoolGen = constantPoolGen;
  }
  
  private void constraintViolated(Instruction i, String message) {
    throw new StaticCodeInstructionOperandConstraintException("Instruction " + Pass3aVerifier.this.tostring(i) + " constraint violated: " + message);
  }
  
  private Method getMethod(JavaClass jc, InvokeInstruction invoke) {
    Method[] ms = jc.getMethods();
    for (Method element : ms) {
      if (element.getName().equals(invoke.getMethodName(this.constantPoolGen)) && 
        Type.getReturnType(element.getSignature()).equals(invoke.getReturnType(this.constantPoolGen)) && 
        Arrays.equals((Object[])Type.getArgumentTypes(element.getSignature()), (Object[])invoke.getArgumentTypes(this.constantPoolGen)))
        return element; 
    } 
    return null;
  }
  
  private Method getMethodRecursive(JavaClass jc, InvokeInstruction invoke) throws ClassNotFoundException {
    Method m = getMethod(jc, invoke);
    if (m != null)
      return m; 
    for (JavaClass superclass : jc.getSuperClasses()) {
      m = getMethod(superclass, invoke);
      if (m != null)
        return m; 
    } 
    for (JavaClass superclass : jc.getInterfaces()) {
      m = getMethod(superclass, invoke);
      if (m != null)
        return m; 
    } 
    return null;
  }
  
  private ObjectType getObjectType(FieldInstruction o) {
    ReferenceType rt = o.getReferenceType(this.constantPoolGen);
    if (rt instanceof ObjectType)
      return (ObjectType)rt; 
    constraintViolated(o, "expecting ObjectType but got " + rt);
    return null;
  }
  
  private void indexValid(Instruction i, int idx) {
    if (idx < 0 || idx >= this.constantPoolGen.getSize())
      constraintViolated(i, "Illegal constant pool index '" + idx + "'."); 
  }
  
  private int maxLocals() {
    try {
      return Repository.lookupClass(Pass3aVerifier.access$000(Pass3aVerifier.this).getClassName()).getMethods()[Pass3aVerifier.access$100(Pass3aVerifier.this)].getCode().getMaxLocals();
    } catch (ClassNotFoundException e) {
      throw new AssertionViolatedException("Missing class: " + e, e);
    } 
  }
  
  public void visitALOAD(ALOAD o) {
    int idx = o.getIndex();
    if (idx < 0) {
      constraintViolated(o, "Index '" + idx + "' must be non-negative.");
    } else {
      int maxminus1 = maxLocals() - 1;
      if (idx > maxminus1)
        constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-1 '" + maxminus1 + "'."); 
    } 
  }
  
  public void visitANEWARRAY(ANEWARRAY o) {
    indexValid(o, o.getIndex());
    Constant c = this.constantPoolGen.getConstant(o.getIndex());
    if (!(c instanceof ConstantClass))
      constraintViolated(o, "Expecting a CONSTANT_Class operand, but found a '" + Pass3aVerifier.this.tostring(c) + "'."); 
    Type t = o.getType(this.constantPoolGen);
    if (t instanceof ArrayType) {
      int dimensions = ((ArrayType)t).getDimensions();
      if (dimensions > 255)
        constraintViolated(o, "Not allowed to create an array with more than 255 dimensions; actual: " + dimensions); 
    } 
  }
  
  public void visitASTORE(ASTORE o) {
    int idx = o.getIndex();
    if (idx < 0) {
      constraintViolated(o, "Index '" + idx + "' must be non-negative.");
    } else {
      int maxminus1 = maxLocals() - 1;
      if (idx > maxminus1)
        constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-1 '" + maxminus1 + "'."); 
    } 
  }
  
  public void visitCHECKCAST(CHECKCAST o) {
    indexValid(o, o.getIndex());
    Constant c = this.constantPoolGen.getConstant(o.getIndex());
    if (!(c instanceof ConstantClass))
      constraintViolated(o, "Expecting a CONSTANT_Class operand, but found a '" + Pass3aVerifier.this.tostring(c) + "'."); 
  }
  
  public void visitDLOAD(DLOAD o) {
    int idx = o.getIndex();
    if (idx < 0) {
      constraintViolated(o, "Index '" + idx + "' must be non-negative. [Constraint by JustIce as an analogon to the single-slot xLOAD/xSTORE instructions; may not happen anyway.]");
    } else {
      int maxminus2 = maxLocals() - 2;
      if (idx > maxminus2)
        constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-2 '" + maxminus2 + "'."); 
    } 
  }
  
  public void visitDSTORE(DSTORE o) {
    int idx = o.getIndex();
    if (idx < 0) {
      constraintViolated(o, "Index '" + idx + "' must be non-negative. [Constraint by JustIce as an analogon to the single-slot xLOAD/xSTORE instructions; may not happen anyway.]");
    } else {
      int maxminus2 = maxLocals() - 2;
      if (idx > maxminus2)
        constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-2 '" + maxminus2 + "'."); 
    } 
  }
  
  public void visitFieldInstruction(FieldInstruction o) {
    try {
      indexValid(o, o.getIndex());
      Constant c = this.constantPoolGen.getConstant(o.getIndex());
      if (!(c instanceof org.apache.bcel.classfile.ConstantFieldref))
        constraintViolated(o, "Indexing a constant that's not a CONSTANT_Fieldref but a '" + Pass3aVerifier.this.tostring(c) + "'."); 
      String fieldName = o.getFieldName(this.constantPoolGen);
      JavaClass jc = Repository.lookupClass(getObjectType(o).getClassName());
      Field[] fields = jc.getFields();
      Field f = null;
      for (Field field : fields) {
        if (field.getName().equals(fieldName)) {
          Type fType = Type.getType(field.getSignature());
          Type oType = o.getType(this.constantPoolGen);
          if (fType.equals(oType)) {
            f = field;
            break;
          } 
        } 
      } 
      if (f == null) {
        JavaClass[] superclasses = jc.getSuperClasses();
        label43: for (JavaClass superclass : superclasses) {
          fields = superclass.getFields();
          for (Field field : fields) {
            if (field.getName().equals(fieldName)) {
              Type fType = Type.getType(field.getSignature());
              Type oType = o.getType(this.constantPoolGen);
              if (fType.equals(oType)) {
                f = field;
                if ((f.getAccessFlags() & 0x5) == 0) {
                  f = null;
                  break label43;
                } 
                break label43;
              } 
            } 
          } 
        } 
        if (f == null)
          constraintViolated(o, "Referenced field '" + fieldName + "' does not exist in class '" + jc.getClassName() + "'."); 
      } else {
        Type.getType(f.getSignature());
        o.getType(this.constantPoolGen);
      } 
    } catch (ClassNotFoundException e) {
      throw new AssertionViolatedException("Missing class: " + e, e);
    } 
  }
  
  public void visitFLOAD(FLOAD o) {
    int idx = o.getIndex();
    if (idx < 0) {
      constraintViolated(o, "Index '" + idx + "' must be non-negative.");
    } else {
      int maxminus1 = maxLocals() - 1;
      if (idx > maxminus1)
        constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-1 '" + maxminus1 + "'."); 
    } 
  }
  
  public void visitFSTORE(FSTORE o) {
    int idx = o.getIndex();
    if (idx < 0) {
      constraintViolated(o, "Index '" + idx + "' must be non-negative.");
    } else {
      int maxminus1 = maxLocals() - 1;
      if (idx > maxminus1)
        constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-1 '" + maxminus1 + "'."); 
    } 
  }
  
  public void visitGETSTATIC(GETSTATIC o) {
    try {
      String fieldName = o.getFieldName(this.constantPoolGen);
      JavaClass jc = Repository.lookupClass(getObjectType(o).getClassName());
      Field[] fields = jc.getFields();
      Field f = null;
      for (Field field : fields) {
        if (field.getName().equals(fieldName)) {
          f = field;
          break;
        } 
      } 
      if (f == null)
        throw new AssertionViolatedException("Field '" + fieldName + "' not found in " + jc.getClassName()); 
      if (!f.isStatic())
        constraintViolated(o, "Referenced field '" + f + "' is not static which it should be."); 
    } catch (ClassNotFoundException e) {
      throw new AssertionViolatedException("Missing class: " + e, e);
    } 
  }
  
  public void visitIINC(IINC o) {
    int idx = o.getIndex();
    if (idx < 0) {
      constraintViolated(o, "Index '" + idx + "' must be non-negative.");
    } else {
      int maxminus1 = maxLocals() - 1;
      if (idx > maxminus1)
        constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-1 '" + maxminus1 + "'."); 
    } 
  }
  
  public void visitILOAD(ILOAD o) {
    int idx = o.getIndex();
    if (idx < 0) {
      constraintViolated(o, "Index '" + idx + "' must be non-negative.");
    } else {
      int maxminus1 = maxLocals() - 1;
      if (idx > maxminus1)
        constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-1 '" + maxminus1 + "'."); 
    } 
  }
  
  public void visitINSTANCEOF(INSTANCEOF o) {
    indexValid(o, o.getIndex());
    Constant c = this.constantPoolGen.getConstant(o.getIndex());
    if (!(c instanceof ConstantClass))
      constraintViolated(o, "Expecting a CONSTANT_Class operand, but found a '" + Pass3aVerifier.this.tostring(c) + "'."); 
  }
  
  public void visitINVOKEDYNAMIC(INVOKEDYNAMIC o) {
    throw new UnsupportedOperationException("INVOKEDYNAMIC instruction is not supported at this time");
  }
  
  public void visitInvokeInstruction(InvokeInstruction o) {
    indexValid(o, o.getIndex());
    if (o instanceof INVOKEVIRTUAL || o instanceof INVOKESPECIAL || o instanceof INVOKESTATIC) {
      Constant c = this.constantPoolGen.getConstant(o.getIndex());
      if (!(c instanceof ConstantMethodref)) {
        constraintViolated(o, "Indexing a constant that's not a CONSTANT_Methodref but a '" + Pass3aVerifier.this.tostring(c) + "'.");
      } else {
        ConstantNameAndType cnat = (ConstantNameAndType)this.constantPoolGen.getConstant(((ConstantMethodref)c).getNameAndTypeIndex());
        ConstantUtf8 cutf8 = (ConstantUtf8)this.constantPoolGen.getConstant(cnat.getNameIndex());
        if (cutf8.getBytes().equals("<init>") && !(o instanceof INVOKESPECIAL))
          constraintViolated(o, "Only INVOKESPECIAL is allowed to invoke instance initialization methods."); 
        if (!cutf8.getBytes().equals("<init>") && cutf8.getBytes().startsWith("<"))
          constraintViolated(o, "No method with a name beginning with '<' other than the instance initialization methods may be called by the method invocation instructions."); 
      } 
    } else {
      Constant c = this.constantPoolGen.getConstant(o.getIndex());
      if (!(c instanceof ConstantInterfaceMethodref))
        constraintViolated(o, "Indexing a constant that's not a CONSTANT_InterfaceMethodref but a '" + Pass3aVerifier.this.tostring(c) + "'."); 
      ConstantNameAndType cnat = (ConstantNameAndType)this.constantPoolGen.getConstant(((ConstantInterfaceMethodref)c).getNameAndTypeIndex());
      String name = ((ConstantUtf8)this.constantPoolGen.getConstant(cnat.getNameIndex())).getBytes();
      if (name.equals("<init>"))
        constraintViolated(o, "Method to invoke must not be '<init>'."); 
      if (name.equals("<clinit>"))
        constraintViolated(o, "Method to invoke must not be '<clinit>'."); 
    } 
    Type t = o.getReturnType(this.constantPoolGen);
    if (t instanceof ArrayType)
      t = ((ArrayType)t).getBasicType(); 
    if (t instanceof ObjectType) {
      Verifier v = VerifierFactory.getVerifier(((ObjectType)t).getClassName());
      VerificationResult vr = v.doPass2();
      if (vr.getStatus() != 1)
        constraintViolated(o, "Return type class/interface could not be verified successfully: '" + vr.getMessage() + "'."); 
    } 
    Type[] ts = o.getArgumentTypes(this.constantPoolGen);
    for (Type element : ts) {
      t = element;
      if (t instanceof ArrayType)
        t = ((ArrayType)t).getBasicType(); 
      if (t instanceof ObjectType) {
        Verifier v = VerifierFactory.getVerifier(((ObjectType)t).getClassName());
        VerificationResult vr = v.doPass2();
        if (vr.getStatus() != 1)
          constraintViolated(o, "Argument type class/interface could not be verified successfully: '" + vr.getMessage() + "'."); 
      } 
    } 
  }
  
  public void visitINVOKEINTERFACE(INVOKEINTERFACE o) {
    try {
      String className = o.getClassName(this.constantPoolGen);
      JavaClass jc = Repository.lookupClass(className);
      Method m = getMethodRecursive(jc, o);
      if (m == null)
        constraintViolated(o, "Referenced method '" + o.getMethodName(this.constantPoolGen) + "' with expected signature '" + o
            .getSignature(this.constantPoolGen) + "' not found in class '" + jc.getClassName() + "'."); 
      if (jc.isClass())
        constraintViolated(o, "Referenced class '" + jc.getClassName() + "' is a class, but not an interface as expected."); 
    } catch (ClassNotFoundException e) {
      throw new AssertionViolatedException("Missing class: " + e, e);
    } 
  }
  
  public void visitINVOKESPECIAL(INVOKESPECIAL o) {
    try {
      String className = o.getClassName(this.constantPoolGen);
      JavaClass jc = Repository.lookupClass(className);
      Method m = getMethodRecursive(jc, o);
      if (m == null)
        constraintViolated(o, "Referenced method '" + o.getMethodName(this.constantPoolGen) + "' with expected signature '" + o
            .getSignature(this.constantPoolGen) + "' not found in class '" + jc.getClassName() + "'."); 
      JavaClass current = Repository.lookupClass(Pass3aVerifier.access$000(Pass3aVerifier.this).getClassName());
      if (current.isSuper() && Repository.instanceOf(current, jc) && !current.equals(jc) && 
        !o.getMethodName(this.constantPoolGen).equals("<init>")) {
        int supidx = -1;
        Method meth = null;
        while (supidx != 0) {
          supidx = current.getSuperclassNameIndex();
          current = Repository.lookupClass(current.getSuperclassName());
          Method[] meths = current.getMethods();
          for (Method meth2 : meths) {
            if (meth2.getName().equals(o.getMethodName(this.constantPoolGen)) && 
              Type.getReturnType(meth2.getSignature()).equals(o.getReturnType(this.constantPoolGen)) && 
              Arrays.equals((Object[])Type.getArgumentTypes(meth2.getSignature()), (Object[])o.getArgumentTypes(this.constantPoolGen))) {
              meth = meth2;
              break;
            } 
          } 
          if (meth != null)
            break; 
        } 
        if (meth == null)
          constraintViolated(o, "ACC_SUPER special lookup procedure not successful: method '" + o.getMethodName(this.constantPoolGen) + "' with proper signature not declared in superclass hierarchy."); 
      } 
    } catch (ClassNotFoundException e) {
      throw new AssertionViolatedException("Missing class: " + e, e);
    } 
  }
  
  public void visitINVOKESTATIC(INVOKESTATIC o) {
    try {
      String className = o.getClassName(this.constantPoolGen);
      JavaClass jc = Repository.lookupClass(className);
      Method m = getMethodRecursive(jc, o);
      if (m == null) {
        constraintViolated(o, "Referenced method '" + o.getMethodName(this.constantPoolGen) + "' with expected signature '" + o
            .getSignature(this.constantPoolGen) + "' not found in class '" + jc.getClassName() + "'.");
      } else if (!m.isStatic()) {
        constraintViolated(o, "Referenced method '" + o.getMethodName(this.constantPoolGen) + "' has ACC_STATIC unset.");
      } 
    } catch (ClassNotFoundException e) {
      throw new AssertionViolatedException("Missing class: " + e, e);
    } 
  }
  
  public void visitINVOKEVIRTUAL(INVOKEVIRTUAL o) {
    try {
      String className = o.getClassName(this.constantPoolGen);
      JavaClass jc = Repository.lookupClass(className);
      Method m = getMethodRecursive(jc, o);
      if (m == null)
        constraintViolated(o, "Referenced method '" + o.getMethodName(this.constantPoolGen) + "' with expected signature '" + o
            .getSignature(this.constantPoolGen) + "' not found in class '" + jc.getClassName() + "'."); 
      if (!jc.isClass())
        constraintViolated(o, "Referenced class '" + jc.getClassName() + "' is an interface, but not a class as expected."); 
    } catch (ClassNotFoundException e) {
      Pass3aVerifier.this.addMessage("Unable to verify INVOKEVITUAL as cannot load target class: " + e.getCause());
    } 
  }
  
  public void visitISTORE(ISTORE o) {
    int idx = o.getIndex();
    if (idx < 0) {
      constraintViolated(o, "Index '" + idx + "' must be non-negative.");
    } else {
      int maxminus1 = maxLocals() - 1;
      if (idx > maxminus1)
        constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-1 '" + maxminus1 + "'."); 
    } 
  }
  
  public void visitLDC(LDC ldc) {
    indexValid(ldc, ldc.getIndex());
    Constant c = this.constantPoolGen.getConstant(ldc.getIndex());
    if (c instanceof ConstantClass) {
      Pass3aVerifier.this.addMessage("Operand of LDC or LDC_W is CONSTANT_Class '" + Pass3aVerifier.this.tostring(c) + "' - this is only supported in JDK 1.5 and higher.");
    } else if (!(c instanceof org.apache.bcel.classfile.ConstantInteger) && !(c instanceof org.apache.bcel.classfile.ConstantFloat) && !(c instanceof org.apache.bcel.classfile.ConstantString)) {
      constraintViolated(ldc, "Operand of LDC or LDC_W must be one of CONSTANT_Integer, CONSTANT_Float or CONSTANT_String, but is '" + Pass3aVerifier.this
          .tostring(c) + "'.");
    } 
  }
  
  public void visitLDC2_W(LDC2_W o) {
    indexValid(o, o.getIndex());
    Constant c = this.constantPoolGen.getConstant(o.getIndex());
    if (!(c instanceof org.apache.bcel.classfile.ConstantLong) && !(c instanceof org.apache.bcel.classfile.ConstantDouble))
      constraintViolated(o, "Operand of LDC2_W must be CONSTANT_Long or CONSTANT_Double, but is '" + Pass3aVerifier.this.tostring(c) + "'."); 
    try {
      indexValid(o, o.getIndex() + 1);
    } catch (StaticCodeInstructionOperandConstraintException e) {
      throw new AssertionViolatedException("Does not BCEL handle that? LDC2_W operand has a problem.", e);
    } 
  }
  
  public void visitLLOAD(LLOAD o) {
    int idx = o.getIndex();
    if (idx < 0) {
      constraintViolated(o, "Index '" + idx + "' must be non-negative. [Constraint by JustIce as an analogon to the single-slot xLOAD/xSTORE instructions; may not happen anyway.]");
    } else {
      int maxminus2 = maxLocals() - 2;
      if (idx > maxminus2)
        constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-2 '" + maxminus2 + "'."); 
    } 
  }
  
  public void visitLoadClass(LoadClass loadClass) {
    ObjectType t = loadClass.getLoadClassType(this.constantPoolGen);
    if (t != null) {
      Verifier v = VerifierFactory.getVerifier(t.getClassName());
      VerificationResult vr = v.doPass1();
      if (vr.getStatus() != 1)
        constraintViolated((Instruction)loadClass, "Class '" + loadClass
            .getLoadClassType(this.constantPoolGen).getClassName() + "' is referenced, but cannot be loaded: '" + vr + "'."); 
    } 
  }
  
  public void visitLOOKUPSWITCH(LOOKUPSWITCH o) {
    int[] matchs = o.getMatchs();
    int max = Integer.MIN_VALUE;
    for (int i = 0; i < matchs.length; i++) {
      if (matchs[i] == max && i != 0)
        constraintViolated(o, "Match '" + matchs[i] + "' occurs more than once."); 
      if (matchs[i] < max) {
        constraintViolated(o, "Lookup table must be sorted but isn't.");
      } else {
        max = matchs[i];
      } 
    } 
  }
  
  public void visitLSTORE(LSTORE o) {
    int idx = o.getIndex();
    if (idx < 0) {
      constraintViolated(o, "Index '" + idx + "' must be non-negative. [Constraint by JustIce as an analogon to the single-slot xLOAD/xSTORE instructions; may not happen anyway.]");
    } else {
      int maxminus2 = maxLocals() - 2;
      if (idx > maxminus2)
        constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-2 '" + maxminus2 + "'."); 
    } 
  }
  
  public void visitMULTIANEWARRAY(MULTIANEWARRAY o) {
    indexValid(o, o.getIndex());
    Constant c = this.constantPoolGen.getConstant(o.getIndex());
    if (!(c instanceof ConstantClass))
      constraintViolated(o, "Expecting a CONSTANT_Class operand, but found a '" + Pass3aVerifier.this.tostring(c) + "'."); 
    int dimensions2create = o.getDimensions();
    if (dimensions2create < 1)
      constraintViolated(o, "Number of dimensions to create must be greater than zero."); 
    Type t = o.getType(this.constantPoolGen);
    if (t instanceof ArrayType) {
      int dimensions = ((ArrayType)t).getDimensions();
      if (dimensions < dimensions2create)
        constraintViolated(o, "Not allowed to create array with more dimensions ('" + dimensions2create + "') than the one referenced by the CONSTANT_Class '" + t + "'."); 
    } else {
      constraintViolated(o, "Expecting a CONSTANT_Class referencing an array type. [Constraint not found in The Java Virtual Machine Specification, Second Edition, 4.8.1]");
    } 
  }
  
  public void visitNEW(NEW o) {
    indexValid(o, o.getIndex());
    Constant c = this.constantPoolGen.getConstant(o.getIndex());
    if (!(c instanceof ConstantClass)) {
      constraintViolated(o, "Expecting a CONSTANT_Class operand, but found a '" + Pass3aVerifier.this.tostring(c) + "'.");
    } else {
      ConstantUtf8 cutf8 = (ConstantUtf8)this.constantPoolGen.getConstant(((ConstantClass)c).getNameIndex());
      Type t = Type.getType("L" + cutf8.getBytes() + ";");
      if (t instanceof ArrayType)
        constraintViolated(o, "NEW must not be used to create an array."); 
    } 
  }
  
  public void visitNEWARRAY(NEWARRAY o) {
    byte t = o.getTypecode();
    if (t != 4 && t != 5 && t != 6 && t != 7 && t != 8 && t != 9 && t != 10 && t != 11)
      constraintViolated(o, "Illegal type code '" + Pass3aVerifier.this.tostring(Byte.valueOf(t)) + "' for 'atype' operand."); 
  }
  
  public void visitPUTSTATIC(PUTSTATIC o) {
    try {
      String fieldName = o.getFieldName(this.constantPoolGen);
      JavaClass jc = Repository.lookupClass(getObjectType(o).getClassName());
      Field[] fields = jc.getFields();
      Field f = null;
      for (Field field : fields) {
        if (field.getName().equals(fieldName)) {
          f = field;
          break;
        } 
      } 
      if (f == null)
        throw new AssertionViolatedException("Field '" + fieldName + "' not found in " + jc.getClassName()); 
      if (f.isFinal() && !Pass3aVerifier.access$000(Pass3aVerifier.this).getClassName().equals(getObjectType(o).getClassName()))
        constraintViolated(o, "Referenced field '" + f + "' is final and must therefore be declared in the current class '" + Pass3aVerifier.access$000(Pass3aVerifier.this).getClassName() + "' which is not the case: it is declared in '" + o
            .getReferenceType(this.constantPoolGen) + "'."); 
      if (!f.isStatic())
        constraintViolated(o, "Referenced field '" + f + "' is not static which it should be."); 
      String methName = Repository.lookupClass(Pass3aVerifier.access$000(Pass3aVerifier.this).getClassName()).getMethods()[Pass3aVerifier.access$100(Pass3aVerifier.this)].getName();
      if (!jc.isClass() && !methName.equals("<clinit>"))
        constraintViolated(o, "Interface field '" + f + "' must be set in a '" + "<clinit>" + "' method."); 
    } catch (ClassNotFoundException e) {
      throw new AssertionViolatedException("Missing class: " + e, e);
    } 
  }
  
  public void visitRET(RET o) {
    int idx = o.getIndex();
    if (idx < 0) {
      constraintViolated(o, "Index '" + idx + "' must be non-negative.");
    } else {
      int maxminus1 = maxLocals() - 1;
      if (idx > maxminus1)
        constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-1 '" + maxminus1 + "'."); 
    } 
  }
  
  public void visitTABLESWITCH(TABLESWITCH o) {}
}
